#ifndef UMAPPP_UMAPPP_HPP
#define UMAPPP_UMAPPP_HPP

/**
 * @file umappp.hpp
 *
 * @brief Overall header for the **umappp** library.
 * This allows for a reasonably intuitive `#include "umappp/umappp.hpp"` call to include the library.
 */

#include "Umap.hpp"

#endif
